unit mstrlist;

{Version 2.1,  11.10.1996, PostCardWare, Albrecht Mengel, mengel@stat-econ.uni-kiel.de
 See mStrList.txt
 If you have problems/ideas with mStrList, please feel free to email me

**** CHANGED:  TMSortType -> TLSortType, soString -> sString, ... , soDate -> sDate  ******
     This was important, otherwise mStrGrid Users should manually include mStrList to their
     USES list, when using setting mStrGrid1.KeyType values. Sorry! (The compiler will warn you)
     If needed you can typecast list.KeyType:=TLSortType(grid.KeyType) or
     grid.KeyType:=TMSortType(list.KeyType). I'll hold TMSortType and TLSortType compatible.

{$R-}

interface

uses classes;

type

  TLSortType = (sString,sStringCaseSensitive,sNumeric,sDate);
  TmStrList = class(TStrings)
  private
    fKeyType:TLSortType;
    fKeyLen:Integer;
    fKeyPos:Integer;
    fScipFirst:Integer;
    FList: PStringItemList;
    FCount: Integer;
    FCapacity: Integer;
    FSorted: Boolean;
    FDuplicates: TDuplicates;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    fSearchSubstring:Boolean;
    fSortDescending:Boolean;
    procedure SetKeyType(value:TLSortType);
    procedure SetKeyLen(value:Integer);
    procedure SetKeyPos(value:Integer);
    procedure ExchangeItems(Index1, Index2: Integer);
    procedure Grow;
    procedure QuickSort(L, R: Integer);
    procedure InsertItem(Index: Integer; const S: string);
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetSorted(Value: Boolean);
    procedure Sort_Alpha;
  protected
    procedure Changed; virtual;
    procedure Changing; virtual;
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure Put(Index: Integer; const S: string); override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
    procedure SetUpdateState(Updating: Boolean); override;
  public
    constructor create;
    destructor Destroy; override;
    function Add(const S: string): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Exchange(Index1, Index2: Integer); override;
    function Find(const S: string; var Index: Integer): Boolean; virtual;
    function IndexOf(const S: string): Integer; override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure Sort; virtual;
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property Sorted: Boolean read FSorted write SetSorted;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
    property KeyType:TLSortType read fKeyType write SetKeyType;
    property KeyLen:Integer read fKeyLen write SetKeyLen;
    property KeyPos:Integer read fKeyPos write SetKeyPos;
    property ScipFirst:Integer read fScipFirst write fScipFirst;
    property SearchSubstring:Boolean read fSearchSubstring write fSearchSubstring;
    property SortDescending:Boolean read fSortDescending write fSortDescending;
  published
  end;

implementation

uses SysUtils, Consts;

FUNCTION rVAL(CONST von:String):Real;
  {Reads a real value out of the string without an error message.
   Returns 0 if no numeric value}
  VAR bis,err:Integer; nach:Real;
  BEGIN VAL(von,nach,bis);
        IF bis>0 THEN VAL(copy(von,1,bis-1),nach,err);
        rVAL:=nach
  END;{rVAL}

constructor TmStrList.create;
begin fKeyPos:=1;
      fKeyLen:=MaxInt;
end;

procedure TmStrList.SetKeyType(value:TLSortType);
begin if value<>fKeyType then
      begin fKeyType:=value;
            if FSorted and (fCount>1)  then
             begin Changing;
                   QuickSort(fScipFirst,fCount-1);
                   if fKeyType=sNumeric then Sort_Alpha;
                   Changed;
end   end    end;

procedure TmStrList.SetKeyLen(value:Integer);
begin if value<>fKeyLen then
      begin if value<1 then value:=1;
            {showmessage('Keylen: '+inttostr(fKeylen)+' -> '+inttostr(value));}
            fKeyLen:=value;
            if FSorted and (fCount>1)  then
             begin Changing;
                   QuickSort(fScipFirst,fCount-1);
                   if fKeyType=sNumeric then Sort_Alpha;
                   Changed;
end   end    end;

procedure TmStrList.SetKeyPos(value:Integer);
begin if value<>fKeyPos then
      begin if value<1 then value:=1;
            fKeyPos:=value;
            if FSorted and (fCount>1) then
             begin Changing;
                   QuickSort(fScipFirst,fCount-1);
                   if fKeyType=sNumeric then Sort_Alpha;
                   Changed;
end   end    end;

procedure TmStrList.QuickSort(L, R: Integer);
var I, J: Integer;
    P: string; Pr:Real; D:TDateTime;
begin
  case fKeyType of
   sString:   repeat I := L; J := R;
                     P := copy(FList^[(L + R) shr 1].FString,fKeyPos,fKeyLen);
                     repeat if fSortDescending
                            then begin while AnsiCompareText(copy(FList^[I].FString,fKeyPos,fKeyLen),P)>0 do Inc(I);
                                       while AnsiCompareText(copy(FList^[J].FString,fKeyPos,fKeyLen),P)<0 do Dec(J);
                                 end
                            else begin while AnsiCompareText(copy(FList^[I].FString,fKeyPos,fKeyLen),P)<0 do Inc(I);
                                       while AnsiCompareText(copy(FList^[J].FString,fKeyPos,fKeyLen),P)>0 do Dec(J);
                                 end;
                            if I <= J then begin
                               ExchangeItems(I, J);
                               Inc(I);      Dec(J);
                            end;
                     until I > J;
                     if L < J then QuickSort(L, J);
                     L := I;
              until I >= R;
   sStringCaseSensitive:
              repeat I := L; J := R;
                     P := copy(FList^[(L + R) shr 1].FString,fKeyPos,fKeyLen);
                     repeat if fSortDescending
                            then begin while copy(FList^[I].FString,fKeyPos,fKeyLen)>P do Inc(I);
                                       while copy(FList^[J].FString,fKeyPos,fKeyLen)<P do Dec(J);
                                 end
                            else begin while copy(FList^[I].FString,fKeyPos,fKeyLen)<P do Inc(I);
                                       while copy(FList^[J].FString,fKeyPos,fKeyLen)>P do Dec(J);
                                 end;
                            if I <= J then begin
                               ExchangeItems(I, J);
                               Inc(I);      Dec(J);
                            end;
                     until I > J;
                     if L < J then QuickSort(L, J);
                     L := I;
              until I >= R;
   sNumeric:  repeat I := L; J := R;
                     Pr:= Rval(copy(FList^[(L + R) shr 1].FString,fKeyPos,fKeyLen));
                     repeat if fSortDescending
                            then begin while Rval(copy(FList^[I].FString,fKeyPos,fKeyLen))>Pr do Inc(I);
                                       while Rval(copy(FList^[J].FString,fKeyPos,fKeyLen))<Pr do Dec(J);
                                 end
                            else begin while Rval(copy(FList^[I].FString,fKeyPos,fKeyLen))<Pr do Inc(I);
                                       while Rval(copy(FList^[J].FString,fKeyPos,fKeyLen))>Pr do Dec(J);
                                 end;
                            if I <= J then begin
                               ExchangeItems(I, J);
                               Inc(I);      Dec(J);
                            end;
                     until I > J;
                     if L < J then QuickSort(L, J);
                     L := I;
              until I >= R;
   sDate:  repeat I := L; J := R;
                     D:= StrToDateTime(copy(FList^[(L + R) shr 1].FString,fKeyPos,fKeyLen));
                     repeat if fSortDescending
                            then begin while StrToDateTime(copy(FList^[I].FString,fKeyPos,fKeyLen))>D do Inc(I);
                                       while StrToDateTime(copy(FList^[J].FString,fKeyPos,fKeyLen))<D do Dec(J);
                                 end
                            else begin while StrToDateTime(copy(FList^[I].FString,fKeyPos,fKeyLen))<D do Inc(I);
                                       while StrToDateTime(copy(FList^[J].FString,fKeyPos,fKeyLen))>D do Dec(J);
                                 end;
                            if I <= J then begin
                               ExchangeItems(I, J);
                               Inc(I);      Dec(J);
                            end;
                     until I > J;
                     if L < J then QuickSort(L, J);
                     L := I;
              until I >= R;
  end
end;

procedure ListError(Ident: Integer);
begin
  raise EListError.CreateRes(Ident);
end;

procedure ListIndexError;
begin
  ListError(SListIndexError);
end;

destructor TmStrList.Destroy;
begin
  FOnChange := nil;
  FOnChanging := nil;
  if FCount <> 0 then Finalize(FList^[0], FCount);
  FCount := 0;
  SetCapacity(0);
end;

function TmStrList.Add(const S: string): Integer;
begin
  if not Sorted then
    Result := FCount
  else
    if Find(S, Result) then
      case Duplicates of
        dupIgnore: Exit;
        dupError: ListError(SDuplicateString);
      end;
  InsertItem(Result, S);
end;

procedure TmStrList.Changed;
begin
  if {!(FUpdateCount = 0) and} Assigned(FOnChange) then FOnChange(Self);
end;

procedure TmStrList.Changing;
begin
  if {!(FUpdateCount = 0) and} Assigned(FOnChanging) then FOnChanging(Self);
end;

procedure TmStrList.Clear;
begin
  if FCount <> 0 then
  begin
    Changing;
    Finalize(FList^[0], FCount);
    FCount := 0;
    SetCapacity(0);
    Changed;
  end;
end;

procedure TmStrList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  Changing;
  Finalize(FList^[Index]);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(TStringItem));
  Changed;
end;

procedure TmStrList.Exchange(Index1, Index2: Integer);
begin
  if (Index1 < 0) or (Index1 >= FCount) or
    (Index2 < 0) or (Index2 >= FCount) then ListIndexError;
  Changing;
  ExchangeItems(Index1, Index2);
  Changed;
end;

procedure TmStrList.ExchangeItems(Index1, Index2: Integer);
var
  Temp: Integer;
  Item1, Item2: PStringItem;
begin
  Item1 := @FList^[Index1];
  Item2 := @FList^[Index2];
  Temp := Integer(Item1^.FString);
  Integer(Item1^.FString) := Integer(Item2^.FString);
  Integer(Item2^.FString) := Temp;
  Temp := Integer(Item1^.FObject);
  Integer(Item1^.FObject) := Integer(Item2^.FObject);
  Integer(Item2^.FObject) := Temp;
end;

function TmStrList.Get(Index: Integer): string;
begin
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  Result := FList^[Index].FString;
end;

function TmStrList.GetCount: Integer;
begin
  Result := FCount;
end;

function TmStrList.GetObject(Index: Integer): TObject;
begin
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  Result := FList^[Index].FObject;
end;

procedure TmStrList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 8 then Delta := 16 else
    if FCapacity > 4 then Delta := 8 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TmStrList.IndexOf(const S: string): Integer;
var i:integer; SearchReal:Real; SearchKey:String; SearchDate:TDateTime;
begin if not Sorted
      then begin SearchReal:=0; SearchDate:=0;
                 {Suchwort konvertieren}
                 case fKeyType of
                   sNumeric:SearchReal:=Rval(S);
                   sString:if fSearchSubstring then SearchKey:=UpperCase(S)
                                               else SearchKey:=S;
                   sDate:SearchDate:=StrToDateTime(S);
                   else SearchKey:=S
                 end;
                 {Suchen}
                 if fSearchSubstring
                 then case fKeyType of
                       sString,
                       sDate    :for i:=0 to FCount-1 do
                                if pos(SearchKey,UpperCase(FList^[i].FString))>0
                                then begin Result:=i; exit end;
                       sStringCaseSensitive:for i:=0 to FCount-1 do
                                if pos(SearchKey,FList^[i].FString)>0
                                then begin Result:=i; exit end;
                       sNumeric:for i:=0 to FCount-1 do
                                if Rval(copy(FList^[i].FString,fKeyPos,fKeyLen))=SearchReal
                                then begin Result:=i; exit end;
                      end
                 else case fKeyType of
                       sString:for i:=0 to FCount-1 do
                                if AnsiCompareText(copy(FList^[i].FString,fKeyPos,fKeyLen),SearchKey)=0
                                then begin Result:=i; exit end;
                       sStringCaseSensitive:for i:=0 to FCount-1 do
                                if copy(FList^[i].FString,fKeyPos,fKeyLen)=SearchKey
                                then begin Result:=i; exit end;
                       sNumeric:for i:=0 to FCount-1 do
                                if Rval(copy(FList^[i].FString,fKeyPos,fKeyLen))=SearchReal
                                then begin Result:=i; exit end;
                       sDate:for i:=0 to FCount-1 do
                                if StrToDateTime(copy(FList^[i].FString,fKeyPos,fKeyLen))=SearchDate
                                then begin Result:=i; exit end;
                      end;
                 {Nicht gefunden}
                 Result:=-1
           end
      else if not Find(S, Result) then Result := -1;
end;

type CmpFunc = function(const S,T:string):integer;

function cmp_string(const S,T:string):integer;
  begin cmp_string:=AnsiCompareText(S,T) end;
function cmp_sens(const S,T:string):integer;
  begin if S<T then cmp_sens:=-1 else
        if S>T then cmp_sens:=1
               else cmp_sens:=0
  end;
function cmp_num(const S,T:string):integer;
  var v,SearchReal:real;
  begin v:=Rval(S);
        SearchReal:=Rval(T);
        if v<SearchReal then cmp_num:=-1 else
        if v>SearchReal then cmp_num:=1
                        else cmp_num:=0
  end;
function cmp_date(const S,T:string):integer;
  var sd,td:TDateTime;
  begin sd:=StrToDateTime(S);
        td:=StrToDateTime(T);
        if sd<td then cmp_date:=-1 else
        if sd>td then cmp_date:=1
                 else cmp_date:=0
  end;

function TmStrList.Find(const S: string; var Index: Integer): Boolean;
var L, H, I, C: Integer; cmp:cmpfunc;
begin {Find}
  {convert searchstring and select compare function}
  case fKeyType of
        sString:cmp:=cmp_string;
        sStringCaseSensitive:cmp:=cmp_sens;
        sDate:cmp:=cmp_date;
        else {soNumeric:}cmp:=cmp_num;
  end;
  Result := False;
  L := 0;
  H := FCount - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := cmp(copy(FList^[I].FString,fKeyPos,fKeyLen),S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        if Duplicates <> dupAccept then L := I;
      end;
    end;
  end;
  Index := L;
end;

procedure TmStrList.Insert(Index: Integer; const S: string);
begin
  if Sorted then ListError(SSortedListError);
  if (Index < 0) or (Index > FCount) then ListIndexError;
  InsertItem(Index, S);
end;

procedure TmStrList.InsertItem(Index: Integer; const S: string);
begin
  Changing;
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(TStringItem));
  with FList^[Index] do
  begin
    Pointer(FString) := nil;
    FObject := nil;
    FString := S;
  end;
  Inc(FCount);
  Changed;
end;

procedure TmStrList.Put(Index: Integer; const S: string);
begin
  if Sorted then ListError(SSortedListError);
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  Changing;
  FList^[Index].FString := S;
  Changed;
end;

procedure TmStrList.PutObject(Index: Integer; AObject: TObject);
begin
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  Changing;
  FList^[Index].FObject := AObject;
  Changed;
end;

procedure TmStrList.SetCapacity(NewCapacity: Integer);
begin
  ReallocMem(FList, NewCapacity * SizeOf(TStringItem));
  FCapacity := NewCapacity;
end;

procedure TmStrList.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then
  begin
    if Value then Sort;
    FSorted := Value;
  end;
end;

procedure TmStrList.SetUpdateState(Updating: Boolean);
begin
  if Updating then Changing else Changed;
end;

procedure TmStrList.Sort;
begin
  if not Sorted and (FCount > 1) then
  begin
    Changing;
    QuickSort(fScipFirst, FCount - 1);
    if fKeyType=sNumeric then Sort_Alpha;
    Changed;
  end;
end;

procedure TmStrList.Sort_Alpha;
var anfang,ende:integer; found:boolean;
begin {soNumeric : post sort Alphas}
      anfang:=fScipFirst;
      found:=false;
      while anfang<fCount do
       if rval(copy(FList^[anfang].FString,fKeyPos,fKeyLen))<>0
        then inc(anfang)
        else begin found:=true;
                   break
             end;
      if not found then exit;
      ende:=anfang+1;
      while ende<fCount do
       if rval(copy(FList^[ende].FString,fKeyPos,fKeyLen))=0
        then inc(ende)
        else break;
      if anfang<ende-1 then
      begin fKeyType:=sString;
            QuickSort(anfang,ende-1);
            fKeyType:=sNumeric
end   end;

end.
